//
//  RCServoController.m
//  RCServo
//
//  Created by Phidgets on 2015-07-27.
//  Copyright (c) 2015 Phidgets. All rights reserved.
//

#import "RCServoController.h"

@implementation RCServoController

- (void)windowWillClose:(NSNotification*)aNotification{
    if(ch)
    {
        //Safely close the example
        Phidget_setOnAttachHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_setOnDetachHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_setOnErrorHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_close((PhidgetHandle)ch);
        PhidgetRCServo_delete(&ch);
        ch = NULL;
    }
}

-(void)awakeFromNib{
    PhidgetReturnCode result;
    
    //Manage GUI
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
    [mainWindow setDelegate:self];
    [phidgetInfoBoxView fillPhidgetInfo:nil];
    [phidgetInfoBox setHidden:NO];
    
    result = PhidgetRCServo_create(&ch);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = [self initChannel:(PhidgetHandle)ch];
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetRCServo_setOnPositionChangeHandler(ch, gotPositionChangeData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetRCServo_setOnVelocityChangeHandler(ch, gotVelocityChangeData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetRCServo_setOnTargetPositionReachedHandler(ch, gotTargetPositionReachedData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    [PhidgetInfoBox openCmdLine:(PhidgetHandle)ch];
}

-(PhidgetReturnCode)initChannel:(PhidgetHandle) channel{
    PhidgetReturnCode result;
    
    result = Phidget_setOnAttachHandler(channel, gotAttach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    result = Phidget_setOnDetachHandler(channel, gotDetach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    result = Phidget_setOnErrorHandler(channel, gotError, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    /*
     * Please review the Phidget22 channel matching documentation for details on the device
     * and class architecture of Phidget22, and how channels are matched to device features.
     */
    
    /*
     * Specifies the serial number of the device to attach to.
     * For VINT devices, this is the hub serial number.
     *
     * The default is any device.
     */
    // Phidget_setDeviceSerialNumber(ch, <YOUR DEVICE SERIAL NUMBER>);
    
    /*
     * For VINT devices, this specifies the port the VINT device must be plugged into.
     *
     * The default is any port.
     */
    // Phidget_setHubPort(ch, 0);
    
    /*
     * Specifies that the channel should only match a VINT hub port.
     * The only valid channel id is 0.
     *
     * The default is 0 (false), meaning VINT hub ports will never match
     */
    // Phidget_setIsHubPortDevice(ch, 1);
    
    /*
     * Specifies which channel to attach to.  It is important that the channel of
     * the device is the same class as the channel that is being opened.
     *
     * The default is any channel.
     */
    // Phidget_setChannel(ch, 0);
    
    /*
     * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
     * In a normal environment this can be done automatically by enabling server discovery, which
     * will cause the client to discovery and connect to available servers.
     *
     * To force the channel to only match a network Phidget, set remote to 1.
     */
    // PhidgetNet_enableServerDiscovery(PHIDGETSERVER_DEVICEREMOTE);
    // Phidget_setIsRemote(ch, 1);
    
    return EPHIDGET_OK;
}

#pragma mark Event callbacks
static void gotAttach(PhidgetHandle phid, void *context){
    [(__bridge id)context performSelectorOnMainThread:@selector(onAttachHandler)
                                           withObject:nil
                                        waitUntilDone:NO];
}
static void gotDetach(PhidgetHandle phid, void *context){
    [(__bridge id)context performSelectorOnMainThread:@selector(onDetachHandler)
                                           withObject:nil
                                        waitUntilDone:NO];
}
static void gotError(PhidgetHandle phid, void *context, Phidget_ErrorEventCode errcode, const char *error){
    [(__bridge id)context performSelectorOnMainThread:@selector(errorHandler:)
                                           withObject:[NSArray arrayWithObjects:[NSNumber numberWithInt:errcode], [NSString stringWithUTF8String:error], nil]
                                        waitUntilDone:NO];
}
static void gotPositionChangeData(PhidgetRCServoHandle phid, void *context, double position){
    [(__bridge id)context performSelectorOnMainThread:@selector(onPositionChangeHandler:)
                                           withObject:[NSNumber numberWithDouble:position]
                                        waitUntilDone:NO];
}
static void gotVelocityChangeData(PhidgetRCServoHandle phid, void *context, double velocity){
    [(__bridge id)context performSelectorOnMainThread:@selector(onVelocityChangeHandler:)
                                           withObject:[NSNumber numberWithDouble:velocity]
                                        waitUntilDone:NO];
}
static void gotTargetPositionReachedData(PhidgetRCServoHandle phid, void *context, double position){
        [(__bridge id)context performSelectorOnMainThread:@selector(onTargetPositionReachedHandler:)
                                               withObject:[NSNumber numberWithDouble:position]
                                            waitUntilDone:NO];
}

#pragma mark Attach, detach, data, and error events
- (void)onAttachHandler{
    PhidgetReturnCode result;
    int engaged, heightChange;
    uint32_t dataInterval, minDataInterval, maxDataInterval;
    double minPulseWidth, maxPulseWidth;
    Phidget_DeviceID deviceID;
    
    //Get information from channel that will allow us to configure the GUI properly
    result = Phidget_getDeviceID((PhidgetHandle)ch, &deviceID);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    result = PhidgetRCServo_getEngaged(ch, &engaged);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    result = PhidgetRCServo_getMinPulseWidth(ch, &minPulseWidth);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    result = PhidgetRCServo_getMaxPulseWidth(ch, &maxPulseWidth);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    //Adjust GUI based on information from channel
    [phidgetInfoBoxView fillPhidgetInfo:(PhidgetHandle)ch];
    [motorCalibrationBox setHidden:NO];
    [motorControlsBox setHidden:NO];
    [motorDataBox setHidden:NO];
    
    if(engaged)
        enabledButton.title = @"Disengage";
    else
        enabledButton.title = @"Engage";

    [minPulseWidthTextField setStringValue:[NSString stringWithFormat:@"%.1f", minPulseWidth]];
    [maxPulseWidthTextField setStringValue:[NSString stringWithFormat:@"%.1f", maxPulseWidth]];
    [positionEventLabel setStringValue:@"unknown"];

    
    [stoppedCheck setHidden:YES];
    switch(deviceID){
        case PHIDID_1066:
        case PHIDID_1061:
            result = PhidgetRCServo_getMinDataInterval(ch, &minDataInterval);
            if(result != EPHIDGET_OK){
                [self outputLastError];
            }
            result = PhidgetRCServo_getMaxDataInterval(ch, &maxDataInterval);
            if(result != EPHIDGET_OK){
                [self outputLastError];
            }
            result = PhidgetRCServo_getDataInterval(ch, &dataInterval);
            if(result != EPHIDGET_OK){
                [self outputLastError];
            }
            [velocityEventLabel setHidden:NO];
            [velocityEventText setHidden:NO];
            [dataIntBox setHidden:NO];
            [dataIntSlider setMinValue:(double)minDataInterval];
            [dataIntSlider setMaxValue:(double)maxDataInterval];
            [dataIntSlider setDoubleValue:(double)dataInterval];
            [dataIntMinLbl setStringValue:[NSString stringWithFormat:@"%d",minDataInterval]];
            [dataIntMaxLbl setStringValue:[NSString stringWithFormat:@"%d",maxDataInterval]];
            [dataIntLbl setStringValue:[NSString stringWithFormat:@"%d ms", dataInterval]];
        case PHIDID_RCC1000:
            [stoppedCheck setHidden:NO];
            heightChange = 525; //used to control window size
            [advancedControlsBox setHidden:NO];
            if(deviceID == PHIDID_RCC1000){
                [voltageCombo setHidden:NO];
                [velocityEventLabel setHidden:YES];
                [velocityEventText setHidden:YES];
            }
            break;
        default:
            [velocityEventLabel setHidden:NO];
            [velocityEventText setHidden:NO];
            if(deviceID == PHIDID_1001)
                [stoppedCheck setHidden:NO];
            heightChange = 330; //used to control window size
            [advancedControlsBox setHidden:YES];
            break;
        
    }
    
    [self updateGUI:YES];
    
    //Adjusting view height
    NSRect frame = [mainWindow frame];
    frame.size.width = 550;
    frame.size.height += heightChange;
    frame.origin.y -= heightChange;
    [mainWindow setFrame:frame display:NO animate:NO];
    [mainWindow center];
}

- (void)onDetachHandler{
    //Reset
    [motorCalibrationBox setHidden:YES];
    [motorControlsBox setHidden:YES];
    [motorDataBox setHidden:YES];
    [advancedControlsBox setHidden:YES];
    [voltageCombo setHidden:YES];
    [velocityEventLabel setHidden:YES];
    [velocityEventText setHidden:YES];
    [dataIntBox setHidden:YES];
    [phidgetInfoBoxView fillPhidgetInfo:nil];
    
    //Adjusting window height
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
}

-(void)onPositionChangeHandler:(NSNumber*)position{
    positionEventLabel.stringValue = [NSString stringWithFormat:@"%.3f", [position doubleValue]];
}

-(void)onVelocityChangeHandler:(NSNumber*)velocity{
    velocityEventLabel.stringValue = [NSString stringWithFormat:@"%.3f", [velocity doubleValue]];
}

- (void)onTargetPositionReachedHandler:(NSNumber *)position{
    [stoppedCheck setIntValue:1];
    positionEventLabel.stringValue = [NSString stringWithFormat:@"%.3f", [position doubleValue]];
}

static int errorCounter = 0;
-(void) outputError:(const char *)errorString{
    errorCounter++;
    NSAttributedString *outputString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[NSString stringWithUTF8String:errorString]]
    	attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
    [[errorEventLog textStorage] beginEditing];
    [[errorEventLog textStorage] appendAttributedString:outputString];
    [[errorEventLog textStorage] endEditing];
    
    [errorEventLogCounter setIntValue:errorCounter];
    if(![errorEventLogWindow isVisible])
        [errorEventLogWindow setIsVisible:YES];
}

-(void)errorHandler:(NSArray *)errorEventData{
    const char* errorString = [[errorEventData objectAtIndex:1] UTF8String];
    [self outputError:errorString];
}

-(void)outputLastError {
	const char *errorString;
	char *errorDetailString = NULL;
    size_t errorDetailStringLen = 0;
    PhidgetReturnCode lastErr;
	
    if (!Phidget_getLastError(&lastErr, &errorString, NULL, &errorDetailStringLen)) {
		errorDetailString = malloc(errorDetailStringLen);
		Phidget_getLastError(&lastErr, &errorString, errorDetailString, &errorDetailStringLen);
    	[self outputError:errorDetailString];
    	free(errorDetailString);
    }
}

- (IBAction)clearErrorLog:(id)sender{
    [[errorEventLog textStorage] setAttributedString:[[NSAttributedString alloc] initWithString:@""]];
    [errorEventLogCounter setIntValue:0];
    errorCounter = 0;
}

- (IBAction)apply:(id)sender{
    PhidgetReturnCode result;

    
    result = PhidgetRCServo_setMinPulseWidth(ch, [minPulseWidthTextField intValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    result = PhidgetRCServo_setMaxPulseWidth(ch, [maxPulseWidthTextField intValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    result = PhidgetRCServo_setMinPosition(ch, [positionAtMinTextField doubleValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    result = PhidgetRCServo_setMaxPosition(ch, [positionAtMaxTextField doubleValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    
    [self updateGUI:NO];
}

- (IBAction)setEnabled:(id)sender{
    PhidgetReturnCode result;
    NSButton *button = (NSButton *)sender;
    
    if(button.state){
        button.title = @"Disengage";
        result = PhidgetRCServo_setEngaged(ch, 1);
        if(result != EPHIDGET_OK){
            [self outputLastError];
        }
    }
    else{
        button.title = @"Engage";
        result = PhidgetRCServo_setEngaged(ch, 0);
        if(result != EPHIDGET_OK){
            [self outputLastError];
        }
    }
}

#pragma mark GUI controls
- (IBAction)setPosition:(id)sender{
    PhidgetReturnCode result;
    int engaged;
    result = PhidgetRCServo_getEngaged(ch, &engaged);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    if(engaged)
        [stoppedCheck setIntValue:0];
    result = PhidgetRCServo_setTargetPosition(ch, [positionSlider doubleValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    [positionText setStringValue:[NSString stringWithFormat:@"%0.3f",[positionSlider doubleValue]]];
}

- (IBAction)setAcceleration:(id)sender{
    PhidgetReturnCode result;
    result = PhidgetRCServo_setAcceleration(ch, [accelerationSlider doubleValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    [accelerationText setStringValue:[NSString stringWithFormat:@"%0.3f",[accelerationSlider doubleValue]]];
}

- (IBAction)setVelocityLimit:(id)sender{
    PhidgetReturnCode result;
    result = PhidgetRCServo_setVelocityLimit(ch, [velocitySlider doubleValue]);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    [velocityText setStringValue:[NSString stringWithFormat:@"%0.3f",[velocitySlider doubleValue]]];
}

- (IBAction)setStopped:(id)sender{
    [stoppedCheck setIntValue:1];
}
- (IBAction)setVoltage:(id)sender{
    PhidgetReturnCode result;
    if([voltageCombo indexOfSelectedItem] == 0)
        result = PhidgetRCServo_setVoltage(ch, RCSERVO_VOLTAGE_5V);
    else if([voltageCombo indexOfSelectedItem] == 1)
        result = PhidgetRCServo_setVoltage(ch, RCSERVO_VOLTAGE_6V);
    else if([voltageCombo indexOfSelectedItem] == 2)
        result = PhidgetRCServo_setVoltage(ch, RCSERVO_VOLTAGE_7_4V);
	else
		result = EPHIDGET_UNEXPECTED;
    if(result != EPHIDGET_OK)
        [self outputLastError];
}
- (IBAction)setSpeedRampingState:(NSButton *)sender {
    PhidgetReturnCode result;
    result = PhidgetRCServo_setSpeedRampingState(ch, (int)sender.state);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    if(sender.state){
        [velocitySlider setEnabled:YES];
        [accelerationSlider setEnabled:YES];
    }
    else{
        [velocitySlider setEnabled:NO];
        [accelerationSlider setEnabled:NO];
    }
}

- (IBAction)setDataInterval:(id)sender{
    PhidgetReturnCode result;
    NSEvent *currentEvent = [[sender window] currentEvent];
    if ([currentEvent type] == NSEventTypeLeftMouseUp) {
        result = PhidgetRCServo_setDataInterval(ch, [dataIntSlider intValue]);
        if(result != EPHIDGET_OK){
            [self outputLastError];
            return;
        }
    }
    [dataIntLbl setStringValue:[NSString stringWithFormat:@"%d ms",[dataIntSlider intValue]]];
}

#pragma mark Misc
-(void)updateGUI:(BOOL)isInitial{
    PhidgetReturnCode result;
    double minVelocityLimit, maxVelocityLimit, velocityLimit;
    double minAcceleration, maxAcceleration, acceleration;
    double minPosition, maxPosition, position;
    Phidget_DeviceID deviceID;
    
    result = Phidget_getDeviceID((PhidgetHandle)ch, &deviceID);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetRCServo_getMinPosition(ch, &minPosition);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    result = PhidgetRCServo_getMaxPosition(ch, &maxPosition);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return;
    }
    
    [positionAtMinTextField setStringValue:[NSString stringWithFormat:@"%.1f", minPosition]];
    [positionAtMaxTextField setStringValue:[NSString stringWithFormat:@"%.1f", maxPosition]];
    [positionMinLabel setStringValue:[NSString stringWithFormat:@"%.3f", minPosition]];
    [positionMaxLabel setStringValue:[NSString stringWithFormat:@"%.3f", maxPosition]];
    if(isInitial){
        [positionText setStringValue:@"unknown"];
        position = minPosition;
    }
    else{
        result = PhidgetRCServo_getPosition(ch, &position);
        if(result != EPHIDGET_OK){
            [self outputLastError];
            return;
        }
        [positionText setStringValue:[NSString stringWithFormat:@"%.3f", position]];
        [positionSlider setDoubleValue:position];
    }
    
    [positionSlider setMinValue:minPosition];
    [positionSlider setMaxValue:maxPosition];
    
    if(deviceID == PHIDID_RCC1000 || deviceID == PHIDID_1061 || deviceID == PHIDID_1066) {
    
		result = PhidgetRCServo_getMinVelocityLimit(ch, &minVelocityLimit);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		result = PhidgetRCServo_getMaxVelocityLimit(ch, &maxVelocityLimit);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		result = PhidgetRCServo_getVelocityLimit(ch, &velocityLimit);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		result = PhidgetRCServo_getMinAcceleration(ch, &minAcceleration);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		result = PhidgetRCServo_getMaxAcceleration(ch, &maxAcceleration);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		result = PhidgetRCServo_getAcceleration(ch, &acceleration);
		if(result != EPHIDGET_OK){
			[self outputLastError];
			return;
		}
		
        [accelerationMinLabel setStringValue:[NSString stringWithFormat:@"%0.3f", minAcceleration]];
        [accelerationMaxLabel setStringValue:[NSString stringWithFormat:@"%0.3f", maxAcceleration]];
        [accelerationSlider setMinValue:minAcceleration];
        [accelerationSlider setMaxValue:maxAcceleration];
        [accelerationSlider setDoubleValue:acceleration];
        [accelerationText setStringValue:[NSString stringWithFormat:@"%0.3f",acceleration]];
        
        [velocityMinLabel setStringValue:[NSString stringWithFormat:@"%0.3f", minVelocityLimit]];
        [velocityMaxLabel setStringValue:[NSString stringWithFormat:@"%0.3f", maxVelocityLimit]];
        [velocitySlider setMinValue:minVelocityLimit];
        [velocitySlider setMaxValue:maxVelocityLimit];
        [velocitySlider setDoubleValue:velocityLimit];
        [velocityText setStringValue:[NSString stringWithFormat:@"%0.3f",velocityLimit]];
    }
}
@end
